/*
 * Copyright (c) 2015 Nokia Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/of_irq.h>
#include <linux/of.h>

#include "linux/cpld-usr.h"
#include "cpld_internal.h"

struct swstat_attribute {
   struct device_attribute dev_attr;
   unsigned bit;

   atomic_t counter;
   struct sysfs_dirent *sd;
   int virq;
};

/* a normal interrupt with a corresponding status bit */
#define _SWSTAT_ATTR(name, bit) \
   { __ATTR(name, 0444, swstat_show, NULL), (bit) }

/* an interrupt without a status bit, or when any interrupt matters whether
 * or not there was a state change */
#define _SWSTAT_COUNT_ATTR(name, bit) \
   { __ATTR(name, 0444, swstat_count_show, NULL), (bit) }


static irqreturn_t cpld_irqnotify_irqthread(int virq, void *data)
{
   struct swstat_attribute *sattr = data;

   atomic_inc(&sattr->counter);
   sysfs_notify_dirent(sattr->sd);

   return IRQ_HANDLED;
}

#if 0
static ssize_t swstat_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct swstat_attribute *sattr = container_of(attr, struct swstat_attribute, dev_attr);
   u8 swstat;
   unsigned bit = sattr->bit;

   if (bit < 8) {
      swstat = readb_cpld_swstat();
   } else {
      swstat = readb_cpld_swstat2();
      bit -= 8;
   }

   return snprintf(buf, PAGE_SIZE, "%u\n", (swstat & (1 << bit)) ? 1 : 0);
}
#endif

static ssize_t swstat_count_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct swstat_attribute *sattr = container_of(attr, struct swstat_attribute, dev_attr);

   return snprintf(buf, PAGE_SIZE, "%u\n", atomic_read(&sattr->counter));
}

static struct swstat_attribute attrs[] = {
   _SWSTAT_COUNT_ATTR(soc2_reset_count, 7),
};

#define NR_IRQNOTIFY ARRAY_SIZE(attrs)

/* dynamically filled in */
static struct attribute *irqnotify_attrs[NR_IRQNOTIFY + 1];

static const struct attribute_group irqnotify_group = {
   .attrs = irqnotify_attrs,
};

int cpld_irqnotify_probe(struct device *dev)
{
   int err;
   unsigned i;
   struct irq_domain *irqdomain;
   const char *name;

   irqdomain = irq_find_host(dev->parent->of_node);
   if (!irqdomain) {
      dev_err(dev, "irq_find_host failed for CPLD\n");
      return -ENODEV;
   }

   for (i = 0; i < NR_IRQNOTIFY; ++i) {
      irqnotify_attrs[i] = &attrs[i].dev_attr.attr;
   }

   err = sysfs_create_group(&dev->kobj, &irqnotify_group);
   if (err)
      return err;

   for (i = 0; i < NR_IRQNOTIFY; ++i) {
      name = irqnotify_attrs[i]->name;

      /* get dirent here to ensure IRQ does not block */
      attrs[i].sd = sysfs_get_dirent(dev->kobj.sd, NULL, name);
      if (!attrs[i].sd) {
         dev_err(dev, "sysfs_get_dirent failed for %u %s\n", i, name);
         err = -ENOMEM;
         goto out;
      }

      attrs[i].virq = irq_create_mapping(irqdomain, attrs[i].bit);
      if (!attrs[i].virq) {
         dev_err(dev, "irq_create_mapping failed for %u %s\n", i, name);
         err = -ENXIO;
         goto out;
      }

      err = request_threaded_irq(attrs[i].virq, NULL, cpld_irqnotify_irqthread, 0, name, &attrs[i]);
      if (err) {
         dev_err(dev, "request_threaded_irq failed for irq %d (%s)\n", attrs[i].virq, name);
         attrs[i].virq = 0;
         goto out;
      }
   }

   return err;

out:
   for (i = 0; i < NR_IRQNOTIFY; ++i) {
      if (attrs[i].virq)
         free_irq(attrs[i].virq, &attrs[i]);
      if (attrs[i].sd)
         sysfs_put(attrs[i].sd);
   }

   sysfs_remove_group(&dev->kobj, &irqnotify_group);


   return err;
}

int cpld_irqnotify_remove(struct device *dev)
{
   unsigned i;

   for (i = 0; i < NR_IRQNOTIFY; ++i) {
      if (attrs[i].virq)
         free_irq(attrs[i].virq, &attrs[i]);
      if (attrs[i].sd)
         sysfs_put(attrs[i].sd);
   }

   sysfs_remove_group(&dev->kobj, &irqnotify_group);

   return 0;
}
