/*
 * Copyright (c) 2014 Nokia Solutions and Networks Inc
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * @file fzm_fpga.h
 * @author Nokia Solutions and Networks
 * @remarks Copyright (C) 2014 Nokia Solutions and Networks
 *
 * @par
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * @par PURPOSE
 * FPGA driver (bus functions, codeload, and irqs)
 *
 * @par USAGE
 * This driver will automatically request firmware through a uevent, which
 * will be handled by udev/mdev.  After the firmware has been loaded, FPGA
 * devices are probed by matching drivers.  The FPGA can be read/written and
 * interrupts used at this point.
 *
 * Modules which use the FPGA should register a fpga driver in their module
 * init code through fpga_driver_register().  They should also register one
 * or more fpga devices using fpga_device_register().  Once the FPGA is
 * available, the driver probe function will be called with the matching
 * device.
 *
 * At this point, the device should be initialized, ensuring to lock the
 * fpga if shared registers are being accessed.  Character device nodes can
 * be added with fpga_cdev_create().  Interrupts on the FPGA can be attached
 * with FPGA_IRQ() and request_threaded_irq().
 *
 * The proper cleanup functions must be called in driver and module remove
 * code.
 *
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    SRID       Description
 * ---------- --------- ---------- ----------------------------------------------
 * 11/23/2010 qdpx86    1372454    Initial creation from mod_cpld
 * 07/29/2011 qdpx86    1426680    Port to ZeNB AP
 * 03/21/2012 qdpx86    1458933    Adding interruptible modify
 * 04/12/2013 yongxu               Port to FZM
 * </pre>
 */
#ifndef FZM_FPGA_CODELOAD_H
#define FZM_FPGA_CODELOAD_H

#include <linux/types.h>
#include <linux/device.h>
#include <linux/fs.h>
#include "fzm_fpga_eis.h"

#ifdef FPGA_DEBUG
#define FUNC_TRACE printk("%s getting called\n", __FUNCTION__)
#else
#define FUNC_TRACE
#endif

struct fpga_device {
   int id;
   //registers, irqs, etc.
   struct device dev;
   char name[];
};

struct fpga_info {
   struct mutex mutex;
};

#define to_fpga_device(x)  container_of((x), struct fpga_device, dev)

struct fpga_driver {
   int (*probe)(struct fpga_device *);
   int (*remove)(struct fpga_device *);
   struct device_driver driver;
};

#define to_fpga_driver(x)  container_of((x), struct fpga_driver, driver)

struct fpga_ioctl_rw {
   uint8_t dev_id;
   uint32_t addr;
   uint32_t mask;
   uint32_t value;
};

#define FPGA_IOCTL_ID 0xE6    // random unused ioctl id
//#define FPGA_READ  _IOWR(FPGA_IOCTL_ID, 0x0, struct fpga_ioctl)
//#define FPGA_WRITE _IOWR(FPGA_IOCTL_ID, 0x1, struct fpga_ioctl)
//#define FPGA_READ  _IOWR(FPGA_IOCTL_ID, 0x0, int)
//#define FPGA_WRITE _IOWR(FPGA_IOCTL_ID, 0x1, int)
#define FPGA_RFIC_READ  0x100
#define FPGA_RFIC_WRITE 0x101
#define FPGA_MASKED_WRITE 0x102

/* internal */
int fpga_ic_probe(struct device *dev);
int fpga_ic_remove(struct device *dev);

/* exported */
struct fpga_device *fpga_device_register(const char *name, int id, void *data, struct device_node *node);
int fpga_driver_register(struct fpga_driver *drv);

int fpga_cdev_create(dev_t devt, const struct file_operations *fops, struct fpga_device *parent, const char *name, ...);
void fpga_cdev_cleanup(struct fpga_device *parent);


/**
 * Lock the FPGA for atomic access.
 * Cannot fail, and will block forever until lock is taken.
 */
void fpga_lock(struct device *dev);

/**
 * Lock the FPGA for atomic access unless a signal is received.
 * Will fail if a signal is received/pending.
 *
 * @return 0 if mutex has been acquired, -errno on failure
 */
int fpga_lock_interruptible(struct device *dev);

/**
 * Attempt to lock the FPGA for atomic access without blocking.
 *
 * @return non-zero if mutex has been acquired
 */
int fpga_trylock(struct device *dev);

/**
 * Unlocks the FPGA to end an atomic acccess.
 */
void fpga_unlock(struct device *dev);

/**
 * Read a register or two from the FPGA without locking.  The fpga_lock MUST
 * already be held.
 *
 * When is_burst is set, a burst read will be issued, which ignores the @c
 * addr parameter (using the latched address from the previous non-burst
 * read), and will read two words into @c val.  Otherwise, one word
 * will be read.
 *
 * @param dev     FPGA device
 * @param addr    address of first register
 * @param val     u32 array to read into (1 or 2 words depending on is_burst)
 * @param dev_id  device to read
 * @param is_burst   false for normal read, true for burst
 *
 * @return 0 on success or -errno on failure
 */
int fpga_read_unlocked(struct device *dev, u32 addr, u32 *val, FPGA_SPI_DEVID_t dev_id, bool is_burst);

/**
 * Write a register or two to the FPGA without locking.  The fpga_lock MUST
 * already be held.
 *
 * When is_burst is set, a burst write will be issued, which ignores the @c
 * addr parameter (using the latched address from the previous non-burst
 * write), and will write two words from @c val.  Otherwise, one word
 * will be written.
 *
 * @param dev     FPGA device
 * @param addr    address of first register
 * @param val     u32 array to write from (1 or 2 words depending on is_burst)
 * @param dev_id  device to write
 * @param is_burst   false for normal write, true for burst
 *
 * @return 0 on success or -errno on failure
 */
int fpga_write_unlocked(struct device *dev, u32 addr, const u32 *val, FPGA_SPI_DEVID_t dev_id, bool is_burst);

/**
 * Read a block of registers from the FPGA without locking.  The fpga_lock
 * MUST already be held.
 *
 * @param dev     FPGA device
 * @param addr    address of first register
 * @param val     u32 array to read into
 * @param dev_id  device to read
 * @param count   number of 32-bit registers to read
 *
 * @return 0 on success or -errno on failure
 */
int fpga_read_block_unlocked(struct device *dev, u32 addr, u32 *val, FPGA_SPI_DEVID_t dev_id, unsigned count);

/**
 * Write a register in the FPGA without locking.  The fpga_lock MUST already
 * be held.
 *
 * @param dev     FPGA device
 * @param addr    address of first register
 * @param val     u32 array to write from
 * @param dev_id  device to write
 * @param count   number of 32-bit registers to write
 *
 * @return 0 on success or -errno on failure
 */
int fpga_write_block_unlocked(struct device *dev, u32 addr, const u32 *val, FPGA_SPI_DEVID_t dev_id, unsigned count);

/**
 * Read a register from the FPGA with locking.
 *
 * @param dev     FPGA device
 * @param addr    address of first register
 * @param val     u32 pointer to read into
 * @param dev_id  device to read
 *
 * @return 0 on success or -errno on failure
 */
int fpga_read(struct device *dev, u32 addr, u32 *val, FPGA_SPI_DEVID_t dev_id);

/**
 * Write a register to the FPGA with locking.
 *
 * @param dev     FPGA device
 * @param addr    address of first register
 * @param val     u32 value to write
 * @param dev_id  device to write
 *
 * @return 0 on success or -errno on failure
 */
int fpga_write(struct device *dev, u32 addr, u32 val, FPGA_SPI_DEVID_t dev_id);


/**
 * Read/modify/write a register in the FPGA without locking.
 * bits in masked bits are set to corresponding value in val
 *
 * @param addr address of register
 * @param mask bits to indicate which bits to modify
 * @param val to provide new bits value indicated by the mask bits
 * @param dev_id to show if it's for fgpa reg or rfic reg
 */
int fpga_modify_unlocked(struct device *dev, u32 addr, u32 mask, u32 val, FPGA_SPI_DEVID_t dev_id);

/**
 * Read/modify/write a register in the FPGA with locking.
 *
 * @param addr address of register
 * @param mask bits to indicate which bits to modify
 * @param val to provide new bits value indicated by the mask bits
 * @param dev_id to show if it's for fgpa reg or rfic reg
 */
int fpga_modify(struct device *dev, u32 addr, u32 mask, u32 val, FPGA_SPI_DEVID_t dev_id);

/**
 * Read/modify/write a register in the FPGA with locking.
 * Will fail if a signal is received/pending.
 *
 * @param addr address of register
 * @param mask bits to preserve from read register
 * @param set bits to set in register
 *
 * @return 0 on success, else -ERESTARTSYS on failure due to signal
 */
static inline int __must_check fpga_modify_interruptible(struct device *dev, u32 addr, u32 mask, u32 val, FPGA_SPI_DEVID_t dev_id)
{
   int err;

   if (fpga_lock_interruptible(dev)) {
      return -ERESTARTSYS;
   }

   err = fpga_modify_unlocked(dev, addr, mask, val, dev_id);

   fpga_unlock(dev);

   return err;
}

/* internal use only */
int fpga_rw(struct device *dev,
            FPGA_SPI_DEVID_t devid,
            FPGA_SPI_MT_t mt,
            u32 pld_0,
            u32 pld_1,
            u32 *data);

#endif
