#ifndef __FPGA_FRAME_H__
#define __FPGA_FRAME_H__

#include <linux/types.h>

#ifndef __KERNEL__
#include <stdbool.h>

typedef struct seqcount {
    unsigned sequence;
} seqcount_t;

/* User-space version of a sequence counter.
 * Once the seq# is even (no writers), read the data, and retry if the
 * seq# changed while we were reading, so we ensure consistent data.
 * Synopsis:
 * do{
 * cur_sn = read_seqcount_begin(&data->seq_num);
 * local_val = data->shared_val;
 * ...
 * }while(read_seqcount_end_retry(&data->seq_num, cur_sn));
 */
static inline unsigned read_seqcount_begin(const seqcount_t *seq)
{
    unsigned prev_seq;
    /* prevent subsequent loads from occuring before the load of sequence */
    while ((prev_seq = __atomic_load_n(&seq->sequence, __ATOMIC_ACQUIRE)) & 1);
    return prev_seq;
}

static inline bool read_seqcount_end_retry(const seqcount_t *seq, unsigned prev_seq)
{
    /* prevent the load of sequence below from being reordered before the
     * load of the data protected by sequence */
    __atomic_thread_fence(__ATOMIC_ACQUIRE);
    return prev_seq != __atomic_load_n(&seq->sequence, __ATOMIC_RELAXED);
}

#else
#include <linux/seqlock.h>
#endif

#include "fzm_ioctls.h"

#define FPGA_FRAME_AIF_NOT_READY_INVALID_SFN   ((__u64)~0ULL)

/*an instance of the following is in shared memory. It's populated by fpga_frame driver and read by ddal_sfn API's*/
struct fpga_frame_info_shm
{
    __u64 sfn; /**< 40-bit sfn */
    struct timespec timestamp;

    /*system tick count. the same thing user space would call to fix the missed irq/jitter if/when we ever wanted to do that. or this could be used for subframe/bcn computation*/
    __u64 sys_counter;

    /*updated whenever interrup handler updates sfn/timestamp...so application can determine if driver is currently writing new sfn/time values to the shared memory region*/
    seqcount_t seq;

    struct timespec timestamp_wall;
};

/*ddal_sfn_wait_for uses this to request fpga_frame driver to wait for a specific frame*/
struct fpga_frame_ioctl_wait_for
{
    __u32 sfn;
};

#endif
